#!/usr/bin/python3 -u
#
# Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
# All rights reserved.
# HARK was developed by researchers in Okuno Laboratory
# at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
#
# Parses javadoc-like comments from C++ files and writes them into well-formed
# (hopefully) XML NodeClass definitions, that can then be used by Overflow to
# extract info about the nodes.
#
# Usage:
# Takes a file as input and prints out the XML, so the read the info from
# to_read.cc and output it to nice_xml.def you would need to do something like:
# 'python info2def.py to_read.cc > nice_xml.def.'

from __future__ import print_function

import sys
import re
import fileinput
import copy
import json
from collections import defaultdict

defs = {'nodes': []}

default_input = {
    'name': "No Name",
    'object_type': "any",
    'description': "No Description Available"
}

default_output = {
    'name': "No Name",
    'object_type': "any",
    'description': "No Description Available"
}

default_param = {
    'name': "No Name",
    'type': "any",
    'value': "",
    'list': [],
    'condition': "",
    'description': "No Description Available"
}

default_node_info = {
    'name': "No Name",
    'category': "No Category",
    'description': "No Description Available",
    'inputs': [],
    'outputs': [],
    'parameters': []
}

node_info = copy.deepcopy(default_node_info)

# the last info parsed, so we can deal with carryovers.
last_info = ""
last_name = ""

# flag so we know if we are in a javadoc comment
in_comment = 0

# assign the inputs and outputs to the default value
current_input  = copy.deepcopy(default_input)
current_output = copy.deepcopy(default_output)
current_param  = copy.deepcopy(default_param)

dep       = defaultdict(list)
filedep   = defaultdict(list)
headerdep = defaultdict(list)

# read through the file
for line in fileinput.input():
    base_filename = fileinput.filename()

    # if we are in a comment, look for the information to parse.
    if in_comment == 1 :

        # look for a '* @something info', then we need to process the item.
        m = re.search("\* (@[a-zA-Z:_0-9]+) +([\<\>a-zA-Z:_0-9 .,;\"\'\/\*\!\+=\-\(\)\[\]\{\}]+)", line)
        if m is not None :
            pname = m.group(1)
            pval  = m.group(2)
            pval = re.sub("\s*$", "", pval, 1)
            pval = re.sub("<", "&lt;", pval)
            pval = re.sub(">", "&gt;", pval)

            if pname == '@name' :
                node_info[pname[1:]] = pval
            elif pname == '@category' :
                node_info[pname[1:]]= pval
            elif pname == '@require' :
                filedep[base_filename].append(pval)
                node_info[pname[1:]] = pval
            elif pname == '@description' :
                node_info[pname[1:]] = pval

            # start of a new input
            elif pname == '@input_name' :
                # check if we have an old input we need to append
                if current_input['name'] != 'No Name' :
                    # add the current input to the inputs
                    node_info['inputs'].append(copy.deepcopy(current_input))
                    # clear the input
                    current_input = copy.deepcopy(default_input)
                current_input['name'] = pval
            elif pname == '@input_type' :
                current_input['object_type'] = pval
            elif pname == '@input_description' :
                current_input['description'] = pval

            # start of a new output
            elif pname == '@output_name' :
                # check if we have an old output we need to append
                if current_output['name'] != 'No Name' :
                    # add the current output to the outputs
                    node_info['outputs'].append(copy.deepcopy(current_output))
                    # clear the info
                    current_output = copy.deepcopy(default_output)
                current_output['name'] = pval
            elif pname == '@output_type' :
                current_output['object_type'] = pval
            elif  pname == '@output_description' :
                current_output['description'] = pval

            # start of a new parameter
            elif  pname == '@parameter_name' :
                # check if we have an old parameter to append to the info
                if  current_param['name'] != 'No Name' :
                    # add the current parameter to the parameter list
                    node_info['parameters'].append(copy.deepcopy(current_param))
                    # clear the parameter info
                    current_param = copy.deepcopy(default_param)
                current_param['name'] = pval
            elif pname == '@parameter_type' :
                current_param['type'] = pval
            elif  pname == '@parameter_value' :
                current_param['value'] = pval
            elif pname == '@parameter_description' :
                current_param['description'] = pval
            elif pname == '@parameter_list' :
                current_param['list'] = pval.split(':')
            elif  pname == '@parameter_valid' :
                current_param['condition'] = pval
            else:
                print("Unrecognized \"@\" tag: {}".format(pname), file=sys.stderr)
                sys.exit()

            # set the added information as the last info added
            last_name = pname[1:]
            last_info = pval

        elif re.search("END\*\/", line) is not None : # look for the trailing '*/' 
            if current_input['name'] != 'No Name' :
                node_info['inputs'].append(copy.deepcopy(current_input))
            if current_output['name'] != 'No Name' :
                node_info['outputs'].append(copy.deepcopy(current_output))
            if current_param['name'] != 'No Name' :
                node_info['parameters'].append(copy.deepcopy(current_param))

            node_info['source'] = base_filename
            defs['nodes'].append(node_info)

            # the last info parsed, so we can deal with carryovers.
            last_info = ""
            last_name = ""

            # flag so we know if we are in a javadoc comment
            in_comment = 0

            # assign the inputs and outputs to the default value
            current_input  = copy.deepcopy(default_input)
            current_output = copy.deepcopy(default_output)
            current_param  = copy.deepcopy(default_param)
            node_info      = copy.deepcopy(default_node_info)
            in_comment     = 0

        else :
            # line spill over from the previous comment. Need to add this
            # info to the previous value
            m = re.search("\* ([\<\>a-zA-Z:_0-9 .,\"\'\//\*\!\+=\-\(\)\[\]\{\}]+)", line)
            if m is not None :
                new_info = last_info + "\n" + m.group(1)
                # if it is input info add it to the current_input
                if last_name[0:5] == 'input' :
                    current_input[last_name] = new_info
                # if it is an output, add it to the current output
                elif last_name[0:6] == 'output' :
                    current_output[last_name] = new_info
                # if it is a param, add it to the current param
                elif last_name[0:9] == 'parameter' :
                    current_param[last_name] = new_info
                # otherwise, we can add it directly to the node info
                else :
                    node_info[last_name] = new_info
                # keep expanding last_info for multiple carryovers
                last_info = new_info
            #elif re.search("\*", line) :
                #print("blank line ")

    # if we aren't in a comment, then keep looking for the beginning
    elif re.search("\/\*Node", line) is not None :
        #print("comment\n")
        in_comment = 1
    else :
        m = re.search("\/\/\@implements ([a-zA-Z0-9_]*)", line)
        if m is not None :
            dep[m.group(1)].append(base_filename)
        else :
            m = re.search("\/\/\@require ([a-zA-Z0-9_]*)", line)
            if m is not None :
                filedep[base_filename].append(m.group(1))
            else :
                m = re.search("^\#include \"([a-zA-Z_0-9\.]*)\"", line)
                if m is not None :
                    headerdep[base_filename].append(m.group(1))

defs['module-depend'] = [
    {'module': module,
     'files':  [{'file': depend} for depend in depends]
    } for module, depends in dep.items()
]

defs['file-depend'] = [
    {'file':    file,
     'modules':
     [] if file not in filedep else
     [{'module': depend} for depend in filedep[file]],
     'headers':
     [] if file not in headerdep else
     [{'header': depend} for depend in headerdep[file]],
    } for file in set(list(filedep.keys()) + list(headerdep.keys()))
]
    
json.dump(defs, sys.stdout, indent=2)

# end of file
